﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class CharacterVisionManager : MonoBehaviour
{
	[Header ("References")]
	[SerializeField]
	private MasterCharacterManager character;
	[SerializeField]
	private Transform visionSourceTransform;
	[SerializeField]
	private MeshFilter meshFilter;

	[Header ("Layer Masks")]
	[SerializeField]
	private LayerMask lightSourceMask;
	[SerializeField]
	private LayerMask visionCollisionMask;
	[SerializeField]
	private LayerMask boundaryWallMask;
	[SerializeField]
	private LayerMask boundaryPointMask;

	[Header ("Vision Varaibles")]
	public float visionRangeDarkness;
	public float visionRangeAttack;
	public float visionRangeMax = 1200;
	public float visionAngle;
	public float lightVisionAngle;

	#region Vision Functions

	public MasterCharacterManager GetClosestVisibleCharacter(LayerMask mask, bool showVisibleRange = false)
	{
		// show what the character can see
		if (showVisibleRange)
		{
			ShowVisibleRange();
		}

		MasterCharacterManager character = null;
		DetectableObject detectedObj = null;
		detectedObj = GetClosestCharacter(null, mask, visionRangeDarkness, visionAngle, false);
		if (detectedObj == null)
		{
			detectedObj = GetClosestCharacter(null, mask, visionRangeAttack, visionAngle, true);
		}

		if (detectedObj != null)
		{
			return detectedObj.GetComponent<MasterCharacterManager>();
		}
		return null;
	}

	private DetectableObject GetClosestCharacter (DetectableObject restrictedObject, LayerMask mask, float characterRange, float angle, bool checkInLight)
	{
		// grab all characters in range
		Collider[] charactersInRange = Physics.OverlapSphere(visionSourceTransform.position, visionRangeDarkness, mask);

		// determine which ones are in field of view
		DetectableObject character = null;
		float charDistance = Mathf.Infinity;
		for (int i = 0; i < charactersInRange.Length; i++)
		{
			// do not select the restricted object
			if (charactersInRange[i] != restrictedObject)
			{
				// determine if this character is closer than the current chosen character
				float newCharDistance = Vector3.Distance(visionSourceTransform.position, charactersInRange[i].transform.position);
				if (newCharDistance < charDistance)
				{
					// check if it is in visible radius
					float targetAngle = Vector3.Angle(visionSourceTransform.forward, charactersInRange[i].transform.position);
					if (targetAngle < angle)
					{
						// now check if it is visible
						Vector3 raycastDir = charactersInRange[i].transform.position - visionSourceTransform.position;
						RaycastHit hit;
						if (Physics.Raycast(visionSourceTransform.position, raycastDir, out hit, visionRangeDarkness, visionCollisionMask))
						{
							if (hit.collider == charactersInRange[i])
							{
								DetectableObject detectedObj = charactersInRange[i].GetComponent<DetectableObject>();
								// now check if they're in light
								if (!checkInLight || detectedObj.IsInLight)
								{
									character = detectedObj;
									charDistance = newCharDistance;
								}
							}
						}
					}
				}
			}
		}

		return character;
	}

	public DetectableObject GetClosestVisibleLightSource(GameObject restrictedObject)
	{
		DetectableObject detectObj = GetClosestLightSource(restrictedObject);
		if (detectObj != null)
		{
			return detectObj;
		}
		return null;
	}

	private DetectableObject GetClosestLightSource (GameObject restrictedObject)
	{
		// grab all active light sources in range
		List<StreetLamp> lamps = MasterGameController.Instance.streetLightController.streetLamps;

		// determine which ones are in field of view
		DetectableObject character = null;
		float charDistance = Mathf.Infinity;
		for (int i = 0; i < lamps.Count; i++)
		{
			// do not select the restricted object and do not select innactive street lamps
			if (lamps[i].gameObject != restrictedObject && lamps[i].IsOn)
			{
				// determine if this light source is closer than the current chosen character
				float newCharDistance = Vector3.Distance(visionSourceTransform.position, lamps[i].transform.position);
				if (newCharDistance < charDistance)
				{
					// check if it is in visible radius
					float targetAngle = Vector3.Angle(visionSourceTransform.forward, lamps[i].transform.position);
					if (targetAngle < lightVisionAngle)
					{
						// now check if it is visible
						Vector3 raycastDir = lamps[i].transform.position - visionSourceTransform.position;
						RaycastHit hit;
						if (Physics.Raycast(visionSourceTransform.position, raycastDir, out hit, visionRangeDarkness, visionCollisionMask))
						{
							if (hit.collider.transform == lamps[i].transform)
							{
								DetectableObject detectedObj = lamps[i].GetComponent<DetectableObject>();
								character = detectedObj;
								charDistance = newCharDistance;
							}
						}
					}
				}
			}
		}

		return character;
	}

	private void ShowVisibleRange()
	{
		Mesh mesh = new Mesh ();
		int[] tris = new int[3]{2, 1 ,0};
		Vector3[] verts = new Vector3[3];

		meshFilter.transform.position = visionSourceTransform.transform.position;
		meshFilter.transform.rotation = visionSourceTransform.transform.rotation;
		verts[0] = Vector3.zero;
		verts[1] = new Vector3(Mathf.Sin((visionAngle) * Mathf.Deg2Rad), 0, 1).normalized * visionRangeDarkness;
		verts[2] = new Vector3 (-verts[1].x, 0, verts[1].z);

		mesh.vertices = verts;
		mesh.triangles = tris;
		mesh.RecalculateNormals();
		meshFilter.mesh = mesh;
	}

	public DetectableObject GetBoundaryLocation(GameObject restrictedObject)
	{
		RaycastHit hit;
		if (Physics.Raycast(visionSourceTransform.position, visionSourceTransform.forward, out hit, Mathf.Infinity, boundaryWallMask))
		{
			Collider[] boundariesInRange = Physics.OverlapSphere(hit.point, 1200, boundaryPointMask);
			GameObject go = new GameObject ();
			go.transform.position = hit.point;

			Transform selectedBoundary = null;
			float shortestDistance = Mathf.Infinity;
			for (int i = 0; i < boundariesInRange.Length; i++)
			{
				if (boundariesInRange[i].gameObject != restrictedObject)
				{
					
					float newCharDistance = Vector3.Distance(visionSourceTransform.position, boundariesInRange[i].transform.position);
					if (newCharDistance < shortestDistance)
					{
						shortestDistance = newCharDistance;
						selectedBoundary = boundariesInRange[i].transform;
					}
				}
			}
			if (selectedBoundary != null)
			{
				return selectedBoundary.GetComponent<DetectableObject>();
			}
		}
		return null;
	}

	#endregion
}
