﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class InfectedCharacterManager : CharacterDecisionManager
{

	// Logic State
	public enum InfectedLogic
	{
		beingInfected,
		searching,
		stalking,
		infecting
	}
	[HideInInspector]
	public InfectedLogic currentInfectedMentality;

	// Targetting
	public DetectableObject currentTarget;
	private bool atDestination;

	// Search Variables
	[SerializeField]
	private float SEARCHING_UPDATE_DURATION = 0.25F;
	private float searchForNewDestinationTimer = 0;

	// Infection Variables
	[SerializeField]
	private float CLOSE_ENOUGH_TO_INFECT = 3.0F;
	[SerializeField]
	private float INFECT_DURATION = 3.0F;
	private float infectTimer = 0;

	#region Initialization Functions

	public override void Initialize()
	{
		base.Initialize();

		// set states
		showVisionCone = true;
		currentInfectedMentality = InfectedLogic.beingInfected;
	}

	#endregion

	#region AI Instruction Functions

	public override void RunUpdate()
	{
		base.RunUpdate();

		if (currentInfectedMentality == InfectedLogic.searching)
		{
			SearchingInstructions();
		}

		if (currentInfectedMentality == InfectedLogic.stalking)
		{
			StalkingInstructions();
		}

		if (currentInfectedMentality == InfectedLogic.infecting)
		{
			InfectingInstructions();
		}
	}

	#endregion

	#region Searching Functions

	public void BeginSearching()
	{
		currentInfectedMentality = InfectedLogic.searching;
		GetNewTargetDestination();
	}

	private void SearchingInstructions()
	{
		// First check if there's any people worthy of being attacked
		MasterCharacterManager prey = characterManager.visionManager.GetClosestVisibleCharacter(characterVisionMask, showVisionCone);

		// If we got prey, then start attacking!
		if (prey != null)
		{
			BeginStalking(prey);
		}
		else if (currentTarget != null)
		{
			// First determine if we reached the destination
			Vector3 pos = transform.position;
			Vector2 flatPos = new Vector2 (pos.x, pos.z);
			if(Vector2.Distance(flatPos, currentTarget.transform.position) < CLOSE_ENOUGH_TO_DEST)
			{
				atDestination = true;
				GetNewTargetDestination();

			}

			// move towards target
			characterManager.GoToPosition(currentTarget.transform.position);
		}
	}

	private void GetNewTargetDestination()
	{
		// don't search for a new target unless we're past the update timer duration
		if (!atDestination && currentTarget != null)
		{
			searchForNewDestinationTimer += Time.deltaTime;
			if (searchForNewDestinationTimer < SEARCHING_UPDATE_DURATION)
			{
				return;
			}
		}

		// reset variables
		searchForNewDestinationTimer = 0;
		atDestination = false;

		// Determine the closest light source that the infected can see
		DetectableObject detObj;
		if (currentTarget != null)
		{
			detObj = characterManager.visionManager.GetClosestVisibleLightSource(currentTarget.gameObject);
		}
		else
		{
			detObj = characterManager.visionManager.GetClosestVisibleLightSource(null);
		}

		if (detObj != null)
		{
			currentTarget = detObj;
			return;
		}

		// Choose a city boundry point that's most straight forward
		GetMostStraightForwardBoundary();
	}

	private void GetMostStraightForwardBoundary()
	{
		DetectableObject dectObj;
		if (currentTarget != null)
		{
			dectObj = characterManager.visionManager.GetBoundaryLocation(currentTarget.gameObject);
		}
		else
		{
			dectObj = characterManager.visionManager.GetBoundaryLocation(null);
		}

		if (dectObj != null)
		{
			currentTarget = dectObj;
		}
		else
		{
			MasterGameController.Instance.SelectRandomBoundaryLocation();
		}
	}

	#endregion

	#region Stalking Functions

	private void BeginStalking(MasterCharacterManager prey)
	{
		currentInfectedMentality = InfectedLogic.stalking;
		currentTarget = prey.GetComponent<DetectableObject>();
	}

	private void StalkingInstructions()
	{
		var pos = transform.position;
		var flatPos = new Vector2 (pos.x, pos.z);
		if (Vector2.Distance(flatPos, currentTarget.transform.position) < CLOSE_ENOUGH_TO_INFECT)
		{
			BeginInfecting();
		}
		else
		{
			// move towards target
			characterManager.GoToPosition(currentTarget.transform.position);
		}
	}

	public void ReleasePrey()
	{
		currentTarget = null;
		BeginSearching();
	}

	#endregion

	#region Infecting Functions

	private void BeginInfecting()
	{
		currentInfectedMentality = InfectedLogic.infecting;
		infectTimer = 0;

		// infect the target
		currentTarget.GetComponent<MasterCharacterManager>().SetInfected();
	}

	private void InfectingInstructions()
	{
		infectTimer += Time.deltaTime;
		if (infectTimer >= INFECT_DURATION)
		{
			currentTarget.GetComponent<MasterCharacterManager>().InfectedSearching();
			ReleasePrey();
		}
	}

	#endregion
}
